'''
Created on 22 Apr 2013

@author: maxz
'''


def checkgrad(f, fprime, x, verbose=True, hyper_names=None, tolerance=1E-1, *args, **kw_args):
    """
    Analytical gradient calculation using a 3-point method
    
    """

    import numpy as np

    # using machine precision to choose h

    eps = np.finfo(float).eps
    step = np.sqrt(eps)  # *max(1E-6, x.min())

    # shake things up a bit by taking random steps for each x dimension
    h = step * np.sign(np.random.uniform(-1, 1, x.shape))

    f_ph = f(x + h, *args, **kw_args)
    f_mh = f(x - h, *args, **kw_args)
    numerical_gradient = (f_ph - f_mh) / (2 * h)
    analytical_gradient_all = fprime(x, *args, **kw_args)

    # ratio = (f_ph - f_mh)/(2*np.dot(h.T, analytical_gradient))

    # if True:
    h = np.zeros_like(x)
    dim = 1
    if f_ph.shape:
        dim = len(f_mh)
    res = np.zeros((len(x), 4, dim))

    if verbose:
        if hyper_names is None:
            hyper_names = map(str, range(len(x)))
        m = max(*[len(h) for h in hyper_names])
        m = "{0!s:%is}" % m
        format_string = m + "  NUM:{1:13.5G}  ANA:{2:13.5G}  RAT:{3: 4.3F}  DIF:{4: 4.3F}"
        # format_string = format_string%(m,'s','!s','f','f', 'f')

    for i in range(len(x)):
        h[i] = step
        f_ph = f(x + h, *args, **kw_args)
        f_mh = f(x - h, *args, **kw_args)

        numerical_gradient = (f_ph - f_mh) / (2 * step)

        # analytical_gradient = fprime(x, *args, **kw_args)[i] ## <---- Exhaustive
        analytical_gradient = analytical_gradient_all[i]

        difference = numerical_gradient - analytical_gradient

        if np.abs(step * analytical_gradient) > 0:
            ratio = (f_ph - f_mh) / (2 * step * analytical_gradient)
        elif np.abs(difference) < tolerance:
            ratio = 1.
        else:
            ratio = 0
        h[i] = 0
        if verbose:
            try:
                if(np.abs(ratio - 1) < tolerance):
                    try:
                        mark = " " + '\033[92m' + u"\u2713" + '\033[0m'  # green checkmark
                    except:
                        mark = ":)"
                else:
                    try:
                        mark = " " + '\033[91m' + u"\u2717" + '\033[0m'  # red crossmark
                    except:
                        mark = "X("
                print format_string.format(hyper_names[i],
                                       numerical_gradient,
                                       analytical_gradient,
                                       ratio,
                                       difference), mark
            except Exception:
                import ipdb;ipdb.set_trace()
        res[i, 0, :] = numerical_gradient
        res[i, 1, :] = analytical_gradient
        res[i, 2, :] = ratio
        res[i, 3, :] = difference

    return res
